import numpy
import logging
import kdcount
import mpsort
import pandas as pd
from six import string_types
import warnings

from nbodykit import CurrentMPIComm
from nbodykit.source.catalog import ArrayCatalog

class CylindricalGroups(object):
    """
    Compute groups of objects using a cylindrical grouping method. We identify
    all satellites within a given cylindrical volume around a central object.

    Results are computed when the object is inititalized, and the result is
    stored in the :attr:`groups` attribute; see the documenation of
    :func:`~CylindricalGroups.run`.

    Input parameters are stored in the :attr:`attrs` attribute dictionary.

    Parameters
    ----------
    source : subclass of :class:`~nbodykit.base.catalog.CatalogSource`
        the input source of particles providing the 'Position' column; the
        grouping algorithm is run on this catalog
    rperp : float
        the radius of the cylinder in the sky plane (i.e., perpendicular
        to the line-of-sight)
    rpar : float
        the radius along the line-of-sight direction; this is 1/2 the
        height of the cylinder
    rankby : str, list, ``None``
        a single or list of column names to rank order the input source by
        before computing the cylindrical groups, such that objects ranked first
        are marked as CGM centrals; if ``None`` is supplied, no sorting will
        be done
    flat_sky_los : bool, optional
        a unit vector of length 3 providing the line-of-sight direction,
        assuming a fixed line-of-sight across the box, e.g., [0,0,1] to use
        the z-axis. If ``None``, the observer at (0,0,0) is used to compute
        the line-of-sight for each pair
    periodic : bool, optional
        whether to use periodic boundary conditions
    BoxSize : float, 3-vector, optional
        the size of the box of the input data; must be provided as
        a keyword or in ``source.attrs`` if ``periodic=True``

    References
    ----------
    Okumura, Teppei, et al. "Reconstruction of halo power spectrum from
    redshift-space galaxy distribution: cylinder-grouping method and halo
    exclusion effect", arXiv:1611.04165, 2016.
    """
    logger = logging.getLogger('CylindricalGroups')

    def __init__(self, source, rankby, rperp, rpar, flat_sky_los=None,
                    periodic=False, BoxSize=None):

        if 'Position' not in source:
            raise ValueError("the 'Position' column must be defined in the input source")

        if rankby is None:
            rankby = []

        if isinstance(rankby, string_types):
            rankby = [rankby]

        for col in rankby:
            if col not in source:
                raise ValueError("cannot rank by column '%s'; no such column" %col)

        self.source = source
        self.comm = source.comm
        self.attrs = {}

        # need BoxSize
        self.attrs['BoxSize'] = numpy.empty(3)
        BoxSize = source.attrs.get('BoxSize', BoxSize)
        if periodic and BoxSize is None:
            raise ValueError("please specify a BoxSize if using periodic boundary conditions")
        self.attrs['BoxSize'][:] = BoxSize

        # LOS must be unit vector
        if flat_sky_los is not None:
            if numpy.isscalar(flat_sky_los) or len(flat_sky_los) != 3:
                raise ValueError("line-of-sight ``flat_sky_los`` should be vector with length 3")
            if not numpy.allclose(numpy.einsum('i,i', flat_sky_los, flat_sky_los), 1.0, rtol=1e-5):
                raise ValueError("line-of-sight ``flat_sky_los`` must be a unit vector")

        # warn if periodic and LOS is None
        if flat_sky_los is None and periodic:
            warnings.warn(("CylindricalGroups using periodic boundary conditions "
                           "with line-of-sight computed from origin (0,0,0); maybe specify a line-of-sight?"))

        # save meta-data
        self.attrs['rpar'] = rpar
        self.attrs['rperp'] = rperp
        self.attrs['periodic'] = periodic
        self.attrs['rankby'] = rankby
        self.attrs['flat_sky_los'] = flat_sky_los

        # log some info
        if self.comm.rank == 0:
            args = (str(rperp), str(rpar))
            self.logger.info("finding groups with rperp=%s and rpar=%s " %args)
            if flat_sky_los is None:
                self.logger.info("  using line-of-sight computed using observer at origin (0,0,0)")
            else:
                self.logger.info("  using line-of-sight vector %s" %str(flat_sky_los))
            msg = "periodic" if periodic else "non-periodic"
            msg = "  using %s boundary conditions" %msg
            if self.attrs['BoxSize'] is not None:
                msg += " (BoxSize = %s)" %str(self.attrs['BoxSize'])
            self.logger.info(msg)


        self.run()

    def run(self):
        """
        Compute the cylindrical groups, saving the results to the
        :attr:`groups` attribute

        Attributes
        ----------
        groups : :class:`~nbodykit.source.catalog.array.ArrayCatalog`
            a catalog holding the result of the grouping. The length of the
            catalog is equal to the length of the input size, i.e., the length
            is equal to the :attr:`size` attribute. The relevant fields are:

            #. cgm_type :
                a flag specifying the type for each object,
                with 0 specifying CGM central and 1 denoting CGM satellite
            #. cgm_haloid :
                The index of the CGM object this object belongs to; an integer
                between 0 and the total number of CGM halos
            #. num_cgm_sats :
                The number of satellites in the CGM halo
        """
        from pmesh.domain import GridND
        from nbodykit.algorithms.fof import split_size_3d

        comm = self.comm
        rperp, rpar = self.attrs['rperp'], self.attrs['rpar']
        rankby = self.attrs['rankby']

        if self.attrs['periodic']:
            boxsize = self.attrs['BoxSize']
        else:
            boxsize = None

        np = split_size_3d(self.comm.size)
        if self.comm.rank == 0:
            self.logger.info("using cpu grid decomposition: %s" %str(np))

        # add a column for original index
        self.source['origind'] = self.source.Index

        # sort the data
        data = self.source.sort(self.attrs['rankby'], usecols=['Position', 'origind'])

        # add a column to track sorted index
        data['sortindex'] = data.Index

        # global min/max across all ranks
        pos = data.compute(data['Position'])
        posmin = numpy.asarray(comm.allgather(pos.min(axis=0))).min(axis=0)
        posmax = numpy.asarray(comm.allgather(pos.max(axis=0))).max(axis=0)

        # domain decomposition
        grid = [
            numpy.linspace(posmin[0], posmax[0], np[0] + 1, endpoint=True),
            numpy.linspace(posmin[0], posmax[1], np[1] + 1, endpoint=True),
            numpy.linspace(posmin[0], posmax[2], np[2] + 1, endpoint=True),
        ]
        domain = GridND(grid, comm=comm)

        # run the CGM algorithm
        groups = cgm(comm, data, domain, rperp, rpar, self.attrs['flat_sky_los'], boxsize)

        # make the final structured array
        self.groups = ArrayCatalog(groups, comm=self.comm, **self.attrs)

        # log some info
        N_cen = (groups['cgm_type']==0).sum()
        isolated_N_cen = ((groups['cgm_type']==0)&(groups['num_cgm_sats']==0)).sum()
        N_cen = self.comm.allreduce(N_cen)
        isolated_N_cen = self.comm.allreduce(isolated_N_cen)
        if self.comm.rank == 0:
            self.logger.info("found %d CGM centrals total" %N_cen)
            self.logger.info("%d/%d are isolated centrals (no satellites)" % (isolated_N_cen,N_cen))

        # delete the column we added to source
        del self.source['origind']

def cgm(comm, data, domain, rperp, rpar, los, boxsize):
    """
    Perform the cylindrical grouping method

    This outputs a structured array with the same length as the input data
    with the following fields for each object in the original data:

    #. cgm_type :
        a flag specifying the type for each object,
        with 0 specifying CGM central and 1 denoting CGM satellite
    #. cgm_haloid :
        The index of the CGM object this object belongs to; an integer
        between 0 and the total number of CGM halos
    #. num_cgm_sats :
        The number of satellites in the CGM halo

    Parameters
    ----------
    comm :
        the MPI communicator
    data : CatalogSource
        catalog with sorted input data, including Position
    domain :
        the domain decomposition
    rperp, rpar : float
        the maximum distances to group objects together in the directions
        perpendicular and parallel to the line-of-sight; the cylinder
        has radius ``rperp`` and height ``2 * rpar``
    los :
        the line-of-sight vector
    boxsize :
        the boxsize, or ``None`` if not using periodic boundary conditions
    """
    # whether we do periodic boundary conditions
    periodic = boxsize is not None
    flat_sky = los is not None

    # the maximum distance still inside the cylinder set by rperp,rpar
    rperp2 = rperp**2; rpar2 = rpar**2
    rmax = (rperp2 + rpar2)**0.5

    pos0, origind0, sortindex0 = data.compute(data['Position'], data['origind'], data['sortindex'])

    layout1    = domain.decompose(pos0, smoothing=0)
    pos1       = layout1.exchange(pos0)
    origind1   = layout1.exchange(origind0)
    sortindex1 = layout1.exchange(sortindex0)

    # exchange particles across ranks, accounting for smoothing radius
    layout2    = domain.decompose(pos1, smoothing=rmax)
    pos2       = layout2.exchange(pos1)
    origind2   = layout2.exchange(origind1)
    sortindex2 = layout2.exchange(sortindex1)
    startrank  = layout2.exchange(numpy.ones(len(pos1), dtype='i4')*comm.rank)

    # make the KD-tree
    tree1 = kdcount.KDTree(pos1, boxsize=boxsize).root
    tree2 = kdcount.KDTree(pos2, boxsize=boxsize).root

    dataframe = []
    j_gt_i = numpy.zeros(len(pos1), dtype='f4')
    wrong_rank = numpy.zeros(len(pos1), dtype='f4')

    def callback(r, i, j):

        r1 = pos1[i]
        r2 = pos2[j]
        dr = r1 - r2

        # enforce periodicity in dpos
        if periodic:
            for axis, col in enumerate(dr.T):
                col[col > boxsize[axis]*0.5] -= boxsize[axis]
                col[col <= -boxsize[axis]*0.5] += boxsize[axis]

        # los distance
        if flat_sky:
            rlos2 =  numpy.einsum("ij,j->i", dr, los)**2
        else:
            center = 0.5 * (r1 + r2)
            dot2 = numpy.einsum('ij, ij->i', dr, center)**2
            center2 = numpy.einsum('ij, ij->i', center, center)
            rlos2 = dot2 / center2

        # sky
        dr2 = numpy.einsum('ij, ij->i', dr, dr)
        rsky2 = numpy.abs(dr2 - rlos2)

        # save the valid pairs
        # To Be Valid: pairs must be within cylinder (compare rperp and rpar)
        valid = (rsky2 <= rperp2)&(rlos2 <= rpar2)
        i = i[valid]; j = j[valid];

        # the correctly sorted indices of particles
        sort_i = sortindex1[i]
        sort_j = sortindex2[j]

        # the rank where the j object lives
        rank_j = startrank[j]

        # track pairs where sorted j > sorted i
        weights = numpy.where(sort_i < sort_j, 1, 0)
        j_gt_i[:] += numpy.bincount(i, weights=weights, minlength=len(pos1))

        # track pairs where j rank is wrong
        weights *= numpy.where(rank_j != comm.rank, 1, 0)
        wrong_rank[:] += numpy.bincount(i, weights=weights, minlength=len(pos1))

        # save the valid pairs for final calculations
        res = numpy.vstack([i, j, sort_i, sort_j]).T
        dataframe.append(res)

    # add all the valid pairs to a dataframe
    tree1.enum(tree2, rmax, process=callback)

    # sorted indices of objects that are centrals
    # (objects with no pairs with j > i)
    centrals = set(sortindex1[(j_gt_i==0)])

    # sorted indices of objects that might be centrals
    # (pairs with j>i that live on other ranks)
    maybes = set(sortindex1[(wrong_rank>0)])

    # store the pairs in a pandas dataframe for fast groupby
    dataframe = numpy.concatenate(dataframe, axis=0)
    df = pd.DataFrame(dataframe, columns=['i', 'j', 'sort_i', 'sort_j'])

    # we sort by the correct sorted index in descending order which puts
    # highest priority objects first
    df.sort_values("sort_i", ascending=False, inplace=True)

    # index by the correct sorted order
    df.set_index('sort_i', inplace=True)

    # to find centrals, considers objects that could be satellites of another
    # (pairs with sort_j > sort_i)
    possible_cens = df[(df['sort_j']>df.index.values)]
    possible_cens = possible_cens.drop(centrals)
    _remove_objects_paired_with(possible_cens, centrals) # remove objs paired with cens

    # sorted indices of objects that have pairs on other ranks
    # these objects are already "maybe" centrals
    on_other_ranks = sortindex1[(wrong_rank>0)]

    # find the centrals and associated halo labels for each central
    all_centrals, labels = _find_centrals(comm, possible_cens, on_other_ranks, centrals, maybes)

    # reset the index and return
    df.reset_index(inplace=True)

    # add the halo labels for each pair in the dataframe
    labels = pd.Series(labels, name='label_i', index=pd.Index(all_centrals, name='sort_i'))
    df = df.join(labels, on='sort_i')
    labels.name = 'label_j'; labels.index.name = 'sort_j'
    df = df.join(labels, on='sort_j')

    # iniitalize the output arrays
    labels = numpy.zeros(len(pos1), dtype='i8') - 1 # indexed by i
    types = numpy.zeros(len(pos1), dtype='u4') # indexed by i
    counts = numpy.zeros(len(pos2), dtype='i8') # indexed by j

    # assign labels of the centrals
    cens = df.dropna(subset=['label_j']).drop_duplicates('i')
    labels[cens['i'].values] = cens['label_i'].values

    # objects on this rank that are satellites
    # (no label for the 1st object in pair but a label for the 2nd object)
    sats = (df['label_i'].isnull())&(~df['label_j'].isnull())
    df = df[sats]

    # find the corresponding central for each satellite
    df = df.sort_values('sort_j', ascending=False)
    df.set_index('sort_i', inplace=True)
    sats_grouped = df.groupby('sort_i', sort=False, as_index=False)
    centrals = sats_grouped.first() # these are the centrals for each satellite

    # update the satellite info with its pair with the highest priority
    cens_i = centrals['i'].values; cens_j = centrals['j'].values
    counts += numpy.bincount(cens_j, minlength=len(pos2))
    types[cens_i] = 1
    labels[cens_i] = centrals['label_j'].values

    # sum counts across ranks (take the sum of any repeated objects)
    counts = layout2.gather(counts, mode='sum')

    # output fields
    dtype = numpy.dtype([('cgm_haloid', 'i8'),
                         ('num_cgm_sats', 'i8'),
                         ('cgm_type', 'u4'),
                         ('origind', 'u4')])
    out = numpy.empty(len(data), dtype=dtype)

    # gather the data back onto the original ranks
    # no ghosts for this domain layout so choose any particle
    out['cgm_haloid'] = layout1.gather(labels, mode='any')
    out['origind'] = layout1.gather(origind1, mode='any')
    out['num_cgm_sats'] = layout1.gather(counts, mode='any')
    out['cgm_type'] = layout1.gather(types, mode='any')

    # restore the original order
    mpsort.sort(out, orderby='origind', comm=comm)

    fields = ['cgm_type', 'cgm_haloid', 'num_cgm_sats']
    return out[fields]


def _remove_objects_paired_with(df, bad_pairs):
    """
    Remove any objects that are paired with an object in ``bad_pairs``

    This is done in place
    """
    assert df.index.name == 'sort_i'

    df.reset_index(inplace=True)
    df.set_index('sort_j', inplace=True)

    # exception could be raised if no pairs need to be dropped
    # so just reset index and return
    try:
        bad_pair_index = df.index.intersection(bad_pairs).unique()
        to_drop = df.loc[bad_pair_index]['sort_i'].values
        df.reset_index(inplace=True)
        df.set_index('sort_i', inplace=True)
        df.drop(to_drop,inplace=True)
    except:
        df.set_index('sort_i', inplace=True)


def _find_centrals(comm, df, on_other_ranks, centrals, maybes):
    """
    Find the sorted index values of all of the centrals

    Each rank determines local centrals (which have no pairs on
    other ranks), and then root searches the objects spread
    out on multiple ranks

    Returns
    -------
    all_centrals : list
        the sorted index values of all centrals
    labels : list
        corresponding labels; ranging from ``0`` to ``len(all_centrals)``
    """
    def find_local_centrals(grp):

        cenid = grp.index.values[0]

        # group number of the centrals that could host this object
        maybe_grp_nums = grp['sort_j'].values

        # this object is a satellite
        if any(num in centrals for num in maybe_grp_nums):
            return

        # this object could be a satellite
        if any(num in maybes for num in maybe_grp_nums):
            maybes.add(cenid)
            return

        # if we get here, this object is definitely a central
        centrals.add(cenid)

    # only need to examine objects that have all higher priority pairs
    # on the same rank --> if they have pairs on other ranks, then they are already
    # marked as "maybes" centrals
    same_rank_df = df.drop(on_other_ranks)

    # group by centrals and find the local centrals
    # these are objects with no higher priority pairs
    cens_grouped = same_rank_df.groupby('sort_i', sort=False)
    cens_grouped.apply(find_local_centrals)

    # the pairs associated with objects that might be satellites
    maybe_index = df.index.intersection(list(maybes)).unique()
    maybe_cen_groups = df.loc[maybe_index]

    # gather data on maybes
    maybes_data = comm.gather(maybe_cen_groups)
    all_centrals = numpy.concatenate(comm.allgather(list(centrals)), axis=0)

    # root identifies the remaining centrals
    if comm.rank == 0:

        # keep track of new centrals
        new_centrals = set()
        all_maybes = pd.concat(maybes_data)

        # remove objects paired with a central
        _remove_objects_paired_with(all_maybes, all_centrals)

        # consider objects in sorted order
        all_maybes.sort_index(ascending=False, inplace=True)

        def finalize(grp):
            grp_nums = grp['sort_j'].values
            if not any(num in new_centrals for num in grp_nums):
                new_centrals.add(grp.index.values[0])

        # find out which of the maybes are actually centrals
        maybes_grouped = all_maybes.groupby('sort_i', sort=False)
        maybes_grouped.apply(finalize)

    else:
        new_centrals = None

    # get the list of all centrals on all ranks
    new_centrals = comm.bcast(new_centrals)
    all_centrals = numpy.append(all_centrals, list(new_centrals))

    # sort and create unique halo labels
    all_centrals[::-1].sort()
    labels = numpy.arange(0, len(all_centrals), dtype='i4')

    return all_centrals, labels

def data_to_sort_key(data):
    """
    Convert floating type data to unique integers for sorting
    """
    return numpy.fromstring(data.tobytes(), dtype='u8')



from .base import FileType
from . import tools
from six import string_types
import numpy
import os
import inspect

class FileStack(FileType):
    """
    A file object that offers a continuous view of a stack of subclasses of
    :class:`~nbodykit.io.base.FileType` instances.

    This allows data to be accessed across multiple files from
    a single file object. The "stack" is a concatenation
    of one file to the end of the previous file.

    Parameters
    ----------
    filetype : subclass of :class:`~nbodykit.io.base.FileType`
        the type of file class to initialize
    path : str
        list of file names, or string specifying single file or
        containing a glob-like '*' pattern
    *args :
        additional arguments to pass to the ``filetype`` instance
        during initialization
    **kwargs :
        additional keyword arguments passed to the ``filetype`` instance
        during initialization
    """
    def __init__(self, filetype, path, *args, **kwargs):

        # check that filetype is subclass of FileType
        if not inspect.isclass(filetype) or not issubclass(filetype, FileType):
            raise ValueError("the stack of `filetype` objects must be subclasses of `FileType`")

        self.path = path

        # save the list of relevant files
        if isinstance(path, list):
            filenames = path
        elif isinstance(path, string_types):
            if '*' in path:
                from glob import glob
                filenames = list(map(os.path.abspath, sorted(glob(path))))
            else:
                if not os.path.exists(path):
                    raise FileNotFoundError(path)
                filenames = [os.path.abspath(path)]
        else:
            raise ValueError("'path' should be a string or a list of strings")
        self.files = [filetype(fn, *args, **kwargs) for fn in filenames]
        self.sizes = numpy.array([len(f) for f in self.files], dtype='i8')

        # set dtype and size
        self.dtype = self.files[0].dtype
        self.size  = self.sizes.sum()

    @property
    def attrs(self):
        """
        Dictionary of meta-data for the stack
        """
        if hasattr(self.files[0], 'attrs'):
            return self.files[0].attrs
        else:
            return {}

    @property
    def nfiles(self):
        """
        The number of files in the FileStack
        """
        return len(self.files)

    def read(self, columns, start, stop, step=1):
        """
        Read the specified column(s) over the given range,
        returning a structured numpy array

        Parameters
        ----------
        columns : str, list of str
            the name of the column(s) to return
        start : int
            the row integer to start reading at
        stop : int
            the row integer to stop reading at
        step : int, optional
            the step size to use when reading; default is 1

        Returns
        -------
        data : array_like
            a numpy structured array holding the requested data
        """
        if isinstance(columns, string_types): columns = [columns]

        toret = []
        for fnum in tools.get_file_slice(self.sizes, start, stop):

            # the local slice
            sl = tools.global_to_local_slice(self.sizes, start, stop, fnum)

            # read this chunk
            toret.append(self.files[fnum].read(columns, sl[0], sl[1], step=1))

        self.logger.debug("Reading column %s [%d:%d] from file %s" % (columns, sl[0], sl[1], self))
        return numpy.concatenate(toret, axis=0)[::step]



from .version import __version__

from mpi4py import MPI

# prevents too many threads exception when using MPI and dask
import dask
dask.set_options(get=dask.get)

_global_options = {}
_global_options['global_cache_size'] = 1e8 # 100 MB
_global_options['dask_chunk_size'] = 100000
_global_options['paint_chunk_size'] = 1024 * 1024 * 8

class CurrentMPIComm(object):
    """
    A class to faciliate getting and setting the current MPI communicator.
    """
    _instance = None

    @staticmethod
    def enable(func):
        """
        Decorator to attach the current MPI communicator to the input
        keyword arguments of ``func``, via the ``comm`` keyword.
        """
        import functools
        @functools.wraps(func)
        def wrapped(*args, **kwargs):
            kwargs.setdefault('comm', None)
            if kwargs['comm'] is None:
                kwargs['comm'] = CurrentMPIComm.get()
            return func(*args, **kwargs)
        return wrapped

    @classmethod
    def get(cls):
        """
        Get the current MPI communicator, returning ``MPI.COMM_WORLD``
        if it has not be explicitly set yet.
        """
        # initialize MPI and set the comm if we need to
        if not cls._instance:
            comm = MPI.COMM_WORLD
            cls._instance = comm

        return cls._instance

    @classmethod
    def set(cls, comm):
        """
        Set the current MPI communicator to the input value.
        """
        cls._instance = comm

class GlobalCache(object):
    """
    A class to faciliate calculation using a global cache via
    :class:`dask.cache.Cache`.
    """
    _instance = None

    @classmethod
    def get(cls):
        """
        Return the global cache object. The default size is controlled
        by the ``global_cache_size`` global option; see :class:`set_options`.

        Returns
        -------
        cache : :class:`dask.cache.Cache`
            the cache object, as provided by dask
        """
        # if not created, use default cache size
        if not cls._instance:
            from dask.cache import Cache
            cls._instance = Cache(_global_options['global_cache_size'])

        return cls._instance

    @classmethod
    def resize(cls, size):
        """
        Re-size the global cache to the specified size in bytes.

        Parameters
        ----------
        size : int, optional
            the desired size of the returned cache in bytes; if not provided,
            the ``global_cache_size`` global option is used
        """
        # get the cachey Cache
        # NOTE: cachey cache stored as the cache attribute of Dask cache
        cache = cls.get().cache

        # set the new size
        cache.available_bytes = size

        # shrink the cache if we need to
        # NOTE: only removes objects if we need to
        cache.shrink()

class set_options(object):
    """
    Set global configuration options.

    Parameters
    ----------
    dask_chunk_size : int
        the number of elements for the default chunk size for dask arrays;
        chunks should usually hold between 10 MB and 100 MB
    global_cache_size : float
        the size of the internal dask cache in bytes; default is 1e9
    paint_chunk_size : int
        the number of objects to paint at the same time. This is independent
        from dask chunksize.
    """
    def __init__(self, **kwargs):
        self.old = _global_options.copy()
        for key in sorted(kwargs):
            if key not in _global_options:
                raise KeyError("Option `%s` is not supported" % key)
        _global_options.update(kwargs)

        # resize the global Cache!
        self.updated_cache_size = False
        if 'global_cache_size' in kwargs:
            GlobalCache.resize(_global_options['global_cache_size'])
            self.updated_cache_size = True

    def __enter__(self):
        return

    def __exit__(self, type, value, traceback):
        _global_options.clear()
        _global_options.update(self.old)

        # resize Cache to original size
        if self.updated_cache_size:
            GlobalCache.resize(_global_options['global_cache_size'])


_logging_handler = None
def setup_logging(log_level="info"):
    """
    Turn on logging, with the specified level.

    Parameters
    ----------
    log_level : 'info', 'debug', 'warning'
        the logging level to set; logging below this level is ignored
    """

    # This gives:
    #
    # [ 000000.43 ]   0: 06-28 14:49  measurestats    INFO     Nproc = [2, 1, 1]
    # [ 000000.43 ]   0: 06-28 14:49  measurestats    INFO     Rmax = 120
    import logging

    levels = {
            "info" : logging.INFO,
            "debug" : logging.DEBUG,
            "warning" : logging.WARNING,
            }

    import time
    logger = logging.getLogger();
    t0 = time.time()

    rank = MPI.COMM_WORLD.rank

    class Formatter(logging.Formatter):
        def format(self, record):
            s1 = ('[ %09.2f ] % 3d: ' % (time.time() - t0, rank))
            return s1 + logging.Formatter.format(self, record)

    fmt = Formatter(fmt='%(asctime)s %(name)-15s %(levelname)-8s %(message)s',
                    datefmt='%m-%d %H:%M ')

    global _logging_handler
    if _logging_handler is None:
        _logging_handler = logging.StreamHandler()
        logger.addHandler(_logging_handler)

    _logging_handler.setFormatter(fmt)
    logger.setLevel(levels[log_level])



import mcfit
import numpy
from scipy.interpolate import InterpolatedUnivariateSpline as spline
from scipy.integrate import quad

from .linear import LinearPower

NUM_PTS = 1024
KMIN = 1e-5
KMAX = 1e2
NMAX = 15

def isiterable(obj):
    """Returns `True` if the given object is iterable."""
    try:
        iter(obj)
        return True
    except TypeError:
        return False

def vectorize_if_needed(func, *x):
    """ Helper function to vectorize functions on array inputs"""
    if any(map(isiterable, x)):
        return numpy.vectorize(func)(*x)
    else:
        return func(*x)

class ZeldovichPower(object):
    """
    The matter power spectrum in the Zel'dovich approximation.

    Parameters
    ----------
    cosmo : :class:`Cosmology`
        the cosmology instance
    z : float
        the redshift of the power spectrum
    transfer : str, optional
        string specifying the transfer function to use for the linear
        power spectrum; one of 'CLASS', 'EisensteinHu', 'NoWiggleEisensteinHu'

    Attributes
    ----------
    cosmo : class:`Cosmology`, astropy.cosmology.FLRW
        the object giving the cosmological parameters
    sigma8 : float
        the z=0 amplitude of matter fluctuations
    redshift : float
        the redshift to compute the power at
    Plin : class:`LinearPower`
        the linear power spectrum class used to compute the Zel'dovich power
    """
    def __init__(self, cosmo, redshift, transfer='CLASS'):

        # initialize the linear power
        self.Plin = LinearPower(cosmo, redshift, transfer='CLASS')

        self.cosmo = self.Plin.cosmo
        self._sigma8 = self.cosmo.sigma8
        self.redshift = redshift

        # use low-k approx below this value
        self._k0_low = 5e-3

        # store meta-data
        self._attrs = {}
        self._attrs.update(self.Plin.attrs)

    @property
    def attrs(self):
        """
        The meta-data dictionary
        """
        self._attrs['redshift'] = self.redshift
        self._attrs['sigma8'] = self.sigma8
        return self._attrs

    def _setup(self):
        r"""
        Internal function to compute the following quantities, needed in
        the Zel'dovich approximation:

        .. math::

            \sigma_v^2 = 1/(6\pi^2) \int dk P_L(k),
            X(q) = \int \frac{dk}{2\pi^2} P_L(k) \left[ \frac{2}{3} - 2\frac{j_1(kq)}{kq} \right],
            Y(q) = \int \frac{dk}{2\pi^2} P_L(k) \left[ -2j_0(kq) + 6\frac{j_1(kq)}{kq} \right].
        """
        # set up the k-grid for integrals
        k = numpy.logspace(numpy.log10(KMIN), numpy.log10(KMAX), NUM_PTS)
        Pk = self.Plin(k)

        # compute the I0, I1 integrals we need
        self._r, I0 = ZeldovichJ0(k)(Pk, extrap=True)
        _, I1 = ZeldovichJ1(k)(Pk, extrap=True)

        # compute the X(r), Y(r) integrals we need
        self._sigmasq = self.Plin.velocity_dispersion(kmin=1e-5, kmax=10., limit=500)**2
        self._X = -2.*I1 + 2 * self._sigmasq
        self._Y = -2.*I0 + 6.*I1

        # needed for the low-k approx
        self._Q3 = quad(lambda q: (self.Plin(q)/q)**2, 1e-6, 100.)[0]

    @property
    def redshift(self):
        """
        The redshift of the power spectrum
        """
        return self._z

    @redshift.setter
    def redshift(self, value):
        self._z = value
        self.Plin.redshift = value
        self._setup()

    @property
    def sigma8(self):
        """
        The amplitude of matter fluctuations at :math:`z=0`.
        """
        return self._sigma8

    @sigma8.setter
    def sigma8(self, value):
        self._sigma8 = value
        self.Plin.sigma8 = value
        self._setup()

    def _low_k_approx(self, k):
        r"""
        Return the low-k approximation of the Zel'dovich power. This computes:

        .. math::
            P(k) = (1 - k^2 \sigma_v^2 + 1/2 k^4 \sigma_v^4) P_L(k) + 0.5 Q_3(k),

        where :math:`Q_3(k)` is defined as

        .. math::
            Q_3(k) = \frac{k^4}{10\pi^2} \int dq \frac{P^2_L(q)}{q^2}.
        """
        Q3 = 1./(10.*numpy.pi**2) * k**4 * self._Q3

        Plin = self.Plin(k)
        term1 = (1 - k**2 * self._sigmasq + 0.5 * k**4 * self._sigmasq**2) * Plin
        term2 = 0.5 * Q3
        return term1 + term2

    def __call__(self, k):
        r"""
        Return the Zel'dovich power in :math:`h^{-3} \mathrm{Mpc}^3 at
        :attr:`redshift` and ``k``, where ``k`` is in units of
        :math:`h \mathrm{Mpc}^{-1}`.

        Parameters
        ----------
        k : float, array_like
            the wavenumbers to evaluate the power at
        """
        def Pzel_at_k(ki):

            # return the low-k approximation
            if ki < self._k0_low:
                return self._low_k_approx(ki)

            # do the full integral
            Pzel = 0.0
            for n in range(0, NMAX+1):

                I = ZeldovichPowerIntegral(self._r, n)
                if n > 0:
                    f = (ki*self._Y)**n * numpy.exp(-0.5*ki**2 * (self._X + self._Y))
                else:
                    f = numpy.exp(-0.5*ki**2 * (self._X + self._Y)) - numpy.exp(-ki**2*self._sigmasq)

                kk, this_Pzel = I(f, extrap=False)
                Pzel += spline(kk, this_Pzel)(ki)

            return Pzel

        return vectorize_if_needed(Pzel_at_k, k)

class ZeldovichJ0(mcfit.mcfit):
    r"""
    An integral over :math:`j_0` needed to compute the Zeldovich power. The
    integral is given by:

    .. math::

        I_0(r) = \int \frac{dk}{2\pi^2} P_L(k) j_0(kr).
    """
    def __init__(self, k):
        self.l = 0
        UK = mcfit.kernels.Mellin_SphericalBesselJ(self.l)
        mcfit.mcfit.__init__(self, k, UK, q=1.0, lowring=False)

        # set pre and post factors
        self.prefac = k
        self.postfac = 1 / (2*numpy.pi)**1.5

class ZeldovichJ1(mcfit.mcfit):
    r"""
    An integral over :math:`j_1` needed to compute the Zeldovich power. The
    integral is given by:

    .. math::

        I_1(r) = \int \frac{dk}{2\pi^2} P_L(k) \frac{j_1(kr)}{kr}.
    """
    def __init__(self, k):
        self.l = 1
        UK = mcfit.kernels.Mellin_SphericalBesselJ(self.l)
        mcfit.mcfit.__init__(self, k, UK, q=0, lowring=False)

        # set pre and post factors
        self.prefac = 1.0
        self.postfac = 1 / (2*numpy.pi)**1.5 / self.y

class ZeldovichPowerIntegral(mcfit.mcfit):
    r"""
    The integral needed to evaluate the density auto spectrum in the
    Zel'dovich approximation.

    This evaluates:

    .. math::
        I(k, n) = 4\pi \int dr r^2 \mathrm{exp}\left[-0.5k^2(X(r) + Y(r)) \right]
                    \left (\frac{k Y(r)}{r} \right)^n j_n(kr).
    """
    def __init__(self, r, n):
        self.n = n
        UK = mcfit.kernels.Mellin_SphericalBesselJ(self.n)
        mcfit.mcfit.__init__(self, r, UK, q=1.5-n, lowring=True)

        # set pre and post factors
        self.prefac = r**(3-n)
        self.postfac = (2*numpy.pi)**1.5



from __future__ import absolute_import
# the future import is important. or in python 2.7 we try to
# import this module itself. Due to the unfortnate name conflict!

from nbodykit.base.mesh import MeshSource
from nbodykit import CurrentMPIComm
from nbodykit.utils import JSONDecoder
from bigfile import BigFileMPI
from pmesh.pm import ParticleMesh, ComplexField, RealField

import numpy
import json
from six import string_types

class BigFileMesh(MeshSource):
    """
    A MeshSource object that reads a mesh from disk using :mod:`bigfile`.

    This can read meshes that have been stored with the
    :func:`~nbodykit.base.mesh.MeshSource.save` function of MeshSource objects.

    Parameters
    ----------
    path : str
        the name of the file to load
    dataset : str
        the name of the dataset in the Bigfile holding the grid
    comm : MPI.Communicator
        the MPI communicator
    **kwargs :
        extra meta-data to be stored in the :attr:`attrs` dict
    """
    def __repr__(self):
        import os
        return "BigFileMesh(file=%s)" % os.path.basename(self.path)

    @CurrentMPIComm.enable
    def __init__(self, path, dataset, comm=None, **kwargs):

        self.path    = path
        self.dataset = dataset
        self.comm    = comm

        # update the meta-data
        self.attrs.update(kwargs)
        with BigFileMPI(comm=self.comm, filename=path)[dataset] as ff:
            for key in ff.attrs:
                v = ff.attrs[key]
                if isinstance(v, string_types) and v.startswith('json://'):
                    self.attrs[key] = json.loads(v[7:], cls=JSONDecoder)
                else:
                    self.attrs[key] = numpy.squeeze(v)

            # fourier space or config space
            if ff.dtype.kind == 'c':
                self.isfourier = True
                if ff.dtype.itemsize == 16:
                    dtype = 'f8'
                else:
                    dtype = 'f4'
            else:
                self.isfourier = False
                if ff.dtype.itemsize == 8:
                    dtype = 'f8'
                else:
                    dtype = 'f4'

        # determine Nmesh
        if 'ndarray.shape' not in self.attrs:
            raise ValueError("`ndarray.shape` should be stored in the Bigfile `attrs` to determine `Nmesh`")

        if 'Nmesh' not in self.attrs:
            raise ValueError("`ndarray.shape` should be stored in the Bigfile `attrs` to determine `Nmesh`")

        Nmesh = self.attrs['Nmesh']
        BoxSize = self.attrs['BoxSize']

        MeshSource.__init__(self, BoxSize=BoxSize, Nmesh=Nmesh, dtype=dtype, comm=comm)

    def to_real_field(self):
        """
        Return the RealField stored on disk.

        .. note::
            The mesh stored on disk must be stored with ``mode=real``

        Returns
        -------
        real : pmesh.pm.RealField
            an array-like object holding the mesh loaded from disk in
            configuration space
        """
        if self.isfourier:
            return NotImplemented

        # the real field to paint to
        pmread = self.pm

        with BigFileMPI(comm=self.comm, filename=self.path)[self.dataset] as ds:
            if self.comm.rank == 0:
                self.logger.info("reading real field from %s" % self.path)
            real2 = RealField(pmread)
            start = sum(self.comm.allgather(real2.size)[:self.comm.rank])
            end = start + real2.size
            real2.unsort(ds[start:end])

        return real2

    def to_complex_field(self):
        """
        Return the ComplexField stored on disk.

        .. note::
            The mesh stored on disk must be stored with ``mode=complex``

        Returns
        -------
        real : pmesh.pm.ComplexField
            an array-like object holding the mesh loaded from disk in Fourier
            space
        """
        if not self.isfourier:
            return NotImplemented
        pmread = self.pm

        if self.comm.rank == 0:
            self.logger.info("reading complex field from %s" % self.path)

        with BigFileMPI(comm=self.comm, filename=self.path)[self.dataset] as ds:
            complex2 = ComplexField(pmread)
            assert self.comm.allreduce(complex2.size) == ds.size
            start = sum(self.comm.allgather(complex2.size)[:self.comm.rank])
            end = start + complex2.size
            complex2.unsort(ds[start:end])

        return complex2



from runtests.mpi import MPITest
from nbodykit.lab import *
from nbodykit import setup_logging
from numpy.testing import assert_array_equal, assert_allclose
import pytest

# debug logging
setup_logging("debug")

@MPITest([4])
def test_tsc_aliasing(comm):

    CurrentMPIComm.set(comm)
    source = UniformCatalog(nbar=3e-2, BoxSize=512., seed=42)
    mesh = source.to_mesh(window='tsc', Nmesh=64, compensated=True)

    # compute the power spectrum -- should be flat shot noise
    # if the compensation worked
    r = FFTPower(mesh, mode='1d', kmin=0.02)
    Pk = r.power['power'].real
    err = (2*Pk**2/r.power['modes'])**0.5

    # test chi2 < 1.0
    residual = (Pk-r.attrs['shotnoise'])/err
    red_chi2 = (residual**2).sum()/len(Pk) # should be about 0.5-0.6
    assert red_chi2 < 1.0

@MPITest([4])
def test_cic_aliasing(comm):

    CurrentMPIComm.set(comm)
    source = UniformCatalog(nbar=3e-2, BoxSize=512., seed=42)
    mesh = source.to_mesh(window='cic', Nmesh=64, compensated=True)

    # compute the power spectrum -- should be flat shot noise
    # if the compensation worked
    r = FFTPower(mesh, mode='1d', kmin=0.02)
    Pk = r.power['power'].real
    err = (2*Pk**2/r.power['modes'])**0.5

    # test chi2 < 1.0
    residual = (Pk-r.attrs['shotnoise'])/err
    red_chi2 = (residual**2).sum()/len(Pk) # should be about 0.5-0.6
    assert red_chi2 < 1.0


@MPITest([1])
def test_fftpower_poles(comm):

    CurrentMPIComm.set(comm)
    source = UniformCatalog(nbar=3e-3, BoxSize=512., seed=42)

    r = FFTPower(source, mode='2d', BoxSize=1024, Nmesh=32, poles=[0,2,4])
    pkmu = r.power['power'].real
    mono = r.poles['power_0'].real

    modes_1d = r.power['modes'].sum(axis=-1)
    mono_from_pkmu = numpy.nansum(pkmu*r.power['modes'], axis=-1) / modes_1d

    assert_array_equal(modes_1d, r.poles['modes'])
    assert_allclose(mono_from_pkmu, mono)


@MPITest([1])
def test_fftpower_padding(comm):

    CurrentMPIComm.set(comm)
    source = UniformCatalog(nbar=3e-3, BoxSize=512., seed=42)

    r = FFTPower(source, mode='1d', BoxSize=1024, Nmesh=32)
    assert r.attrs['N1'] != 0
    assert r.attrs['N2'] != 0

@MPITest([1])
def test_fftpower_padding(comm):

    CurrentMPIComm.set(comm)
    source = UniformCatalog(nbar=3e-3, BoxSize=512., seed=42)

    r = FFTPower(source, mode='1d', BoxSize=1024, Nmesh=32)
    assert r.attrs['N1'] != 0
    assert r.attrs['N2'] != 0

@MPITest([1])
def test_fftpower_save(comm):

    CurrentMPIComm.set(comm)
    source = UniformCatalog(nbar=3e-3, BoxSize=512., seed=42)

    r = FFTPower(source, mode='2d', Nmesh=32)
    r.save('fftpower-test.json')

    r2 = FFTPower.load('fftpower-test.json')

    assert_array_equal(r.power['k'], r2.power['k'])
    assert_array_equal(r.power['power'], r2.power['power'])
    assert_array_equal(r.power['mu'], r2.power['mu'])
    assert_array_equal(r.power['modes'], r2.power['modes'])

@MPITest([1])
def test_fftpower(comm):

    CurrentMPIComm.set(comm)
    source = UniformCatalog(nbar=3e-3, BoxSize=512., seed=42)

    r = FFTPower(source, mode='1d', Nmesh=32)
    # the zero mode is cleared
    assert_array_equal(r.power['power'][0], 0)

@MPITest([1])
def test_fftpower_mismatch_boxsize(comm):

    cosmo = cosmology.Planck15
    CurrentMPIComm.set(comm)

    # input sources
    source1 = UniformCatalog(nbar=3e-3, BoxSize=512., seed=42)
    Plin = cosmology.LinearPower(cosmo, 0.55, transfer='NoWiggleEisensteinHu')
    source2 = LinearMesh(Plin, BoxSize=1024, Nmesh=32, seed=33)

    r = FFTPower(source1, second=source2, mode='1d', BoxSize=1024, Nmesh=32)

@MPITest([1])
def test_fftpower_mismatch_boxsize_fail(comm):

    cosmo = cosmology.Planck15
    CurrentMPIComm.set(comm)

    # input sources
    mesh1 = UniformCatalog(nbar=3e-3, BoxSize=512., seed=42).to_mesh(Nmesh=32)
    Plin = cosmology.LinearPower(cosmo, 0.55, transfer='NoWiggleEisensteinHu')
    mesh2 = LinearMesh(Plin, BoxSize=1024, Nmesh=32, seed=33)

    # raises an exception b/c meshes have different box sizes
    with pytest.raises(ValueError):
        r = FFTPower(mesh1, second=mesh2, mode='1d', BoxSize=1024, Nmesh=32)

@MPITest([1])
def test_projectedpower(comm):

    CurrentMPIComm.set(comm)
    source = UniformCatalog(nbar=3e-3, BoxSize=512., seed=42)

    Nmesh = 64
    rp1 = ProjectedFFTPower(source, Nmesh=Nmesh, axes=[1])

    # the zero mode is cleared
    assert_array_equal(rp1.power['power'][0], 0)

    rp2 = ProjectedFFTPower(source, Nmesh=Nmesh, axes=[0, 1])
    # the zero mode is cleared
    assert_array_equal(rp2.power['power'][0], 0)

    rf = FFTPower(source, Nmesh=Nmesh, mode='1d')

    # FIXME: why a factor of 2?
    assert_allclose(rp1.power['power'][1:].mean() * source.attrs['BoxSize'][0] ** 2, rf.power['power'][1:].mean(), rtol=2 * (Nmesh / 2)**-0.5)
    assert_allclose(rp2.power['power'][1:].mean() * source.attrs['BoxSize'][0], rf.power['power'][1:].mean(), rtol=2 * (Nmesh ** 2 / 2)**-0.5)



from runtests.mpi import MPITest
from nbodykit.lab import *
from nbodykit import setup_logging

from numpy.testing import assert_allclose

setup_logging("debug")

@MPITest([4])
def test_lognormal_sparse(comm):
    cosmo = cosmology.Planck15
    CurrentMPIComm.set(comm)

    # this should generate 15 particles
    Plin = cosmology.LinearPower(cosmo, redshift=0.55, transfer='EisensteinHu')
    source = LogNormalCatalog(Plin=Plin, nbar=1e-5, BoxSize=128., Nmesh=8, seed=42)

    mesh = source.to_mesh(compensated=False)

    real = mesh.paint(mode='real')
    assert_allclose(real.cmean(), 1.0)

@MPITest([1, 4])
def test_lognormal_dense(comm):
    cosmo = cosmology.Planck15
    CurrentMPIComm.set(comm)

    Plin = cosmology.LinearPower(cosmo, redshift=0.55, transfer='EisensteinHu')
    source = LogNormalCatalog(Plin=Plin, nbar=0.2e-2, BoxSize=128., Nmesh=8, seed=42)
    mesh = source.to_mesh(compensated=False)

    real = mesh.paint(mode='real')
    assert_allclose(real.cmean(), 1.0, rtol=1e-5)

@MPITest([1])
def test_lognormal_velocity(comm):
    cosmo = cosmology.Planck15
    CurrentMPIComm.set(comm)

    Plin = cosmology.LinearPower(cosmo, redshift=0.55, transfer='EisensteinHu')
    source = LogNormalCatalog(Plin=Plin, nbar=0.5e-2, BoxSize=1024., Nmesh=32, seed=42)

    source['Value'] = source['Velocity'][:, 0]**2
    mesh = source.to_mesh(compensated=False)

    real = mesh.paint(mode='real')
    velsum = comm.allreduce((source['Velocity'][:, 0]**2).sum().compute())
    velmean = velsum / source.csize

    assert_allclose(real.cmean(), velmean, rtol=1e-5)



from runtests.mpi import MPITest
from nbodykit.lab import *
from nbodykit import setup_logging
from numpy.testing import assert_allclose, assert_array_equal
import pytest

setup_logging("debug")

@MPITest([1, 4])
def test_table(comm):

    from astropy.table import Table
    CurrentMPIComm.set(comm)

    data = numpy.ones(100, dtype=[
            ('Position', ('f4', 3)),
            ('Velocity', ('f4', 3))]
            )
    data = Table(data)
    source = ArrayCatalog(data, BoxSize=100, Nmesh=32)

    for col in ['Position', 'Velocity']:
        assert_array_equal(data[col], source[col])


@MPITest([1, 4])
def test_nonstructured_input(comm):

    from astropy.table import Table
    CurrentMPIComm.set(comm)

    # data should be structured!
    data = numpy.ones(100)
    with pytest.raises(ValueError):
        source = ArrayCatalog(data, BoxSize=100, Nmesh=32)




@MPITest([1, 4])
def test_array(comm):

    cosmo = cosmology.Planck15
    CurrentMPIComm.set(comm)
    data = numpy.ones(100, dtype=[
            ('Position', ('f4', 3)),
            ('Velocity', ('f4', 3))]
            )
    source = ArrayCatalog(data, BoxSize=100, Nmesh=32)

    assert source.csize == 100 * comm.size
    source['Velocity'] = source['Position'] + source['Velocity']
    source['Position'] = source['Position'] + source['Velocity']

    # Position triggers  Velocity which triggers Position and Velocity
    # which resolves to the true data.
    # so total is 3.
    assert_allclose(source['Position'], 3)


@MPITest([1, 4])
def test_dict(comm):

    cosmo = cosmology.Planck15
    CurrentMPIComm.set(comm)
    data = numpy.ones(100, dtype=[
            ('Position', ('f4', 3)),
            ('Velocity', ('f4', 3))]
            )
    # use a dictionary
    data = dict(Position=data['Position'], Velocity=data['Velocity'])
    source = ArrayCatalog(data, BoxSize=100, Nmesh=32)

    assert source.csize == 100 * comm.size
    source['Velocity'] = source['Position'] + source['Velocity']
    source['Position'] = source['Position'] + source['Velocity']
    assert_allclose(source['Position'], 3)



from runtests.mpi import MPITest
from nbodykit.lab import *
from nbodykit import setup_logging, set_options, GlobalCache

# debug logging
setup_logging("debug")

@MPITest([1, 4])
def test_fftpower(comm):
    cosmo = cosmology.Planck15

    CurrentMPIComm.set(comm)

    # lognormal particles
    Plin = cosmology.LinearPower(cosmo, redshift=0.55, transfer='EisensteinHu')
    source = LogNormalCatalog(Plin=Plin, nbar=3e-7, BoxSize=1380., Nmesh=8, seed=42)

    # apply RSD
    source['Position'] += source['VelocityOffset'] * [0,0,1]

    # compute P(k,mu) and multipoles
    result = FFTPower(source, mode='2d', poles=[0,2,4], los=[0,0,1])

    # and save
    output = "./test_fftpower-%d.json" % comm.size
    result.save(output)

@MPITest([1, 4])
def test_paint(comm):
    cosmo = cosmology.Planck15

    CurrentMPIComm.set(comm)

    # lognormal particles
    Plin = cosmology.LinearPower(cosmo, redshift=0.55, transfer='EisensteinHu')
    source = LogNormalCatalog(Plin=Plin, nbar=3e-7, BoxSize=1380., Nmesh=8, seed=42)

    # apply RSD
    source['Position'] += source['VelocityOffset'] * [0,0,1]

    # convert to mesh, with Painter specifics
    source = source.to_mesh(Nmesh=64, BoxSize=1380., interlaced=True, window='tsc', compensated=True)

    def filter(k, v):
        kk = sum(ki ** 2 for ki in k)
        kk[kk == 0] = 1
        return v / kk

    source = source.apply(filter)

    real = source.paint(mode='real')
    complex = source.paint(mode='complex')

    source.save(output="./test_paint-real-%d.bigfile" % comm.size, mode='real')
    source.save(output="./test_paint-complex-%d.bigfile" % comm.size, mode='complex')

@MPITest([1, 4])
def test_set_options(comm):

    CurrentMPIComm.set(comm)

    with set_options(global_cache_size=5e9, dask_chunk_size=75):
        s = UniformCatalog(1000, 1.0)

        # check cache size
        cache = GlobalCache.get()
        assert cache.cache.available_bytes == 5e9

        # check chunk size
        assert s['Position'].chunks[0][0] == 75

    s = UniformCatalog(1000, 1.0)
    assert s['Position'].chunks[0][0] == s.size
